/*-----------------------------------------------------------------------------
 *  imond-stat.c   - send a status message to imond
 *
 *  usage:	imond-stat message
 *  example:	imond-stat "pppoe: up"
 *
 *  Copyright (c) 2000-2001 Frank Meyer <frank@fli4l.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Creation:	    26.11.2000  fm
 *  Last Update:    09.09.2001  fm
 *-----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/errno.h>
#include <string.h>

#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>				/* decl of inet_addr()	    */
#include <sys/socket.h>

#define TRUE		1
#define FALSE		0

/*----------------------------------------------------------------------------
 *  service_connect (host_name, port)	    - connect to tcp-service
 *----------------------------------------------------------------------------
 */
static int
service_connect (char * host_name, int port)
{
    struct sockaddr_in	addr;
    struct hostent *	host_p;
    int			fd;
    int			opt = 1;

    (void) memset ((char *) &addr, 0, sizeof (addr));

    if ((addr.sin_addr.s_addr = inet_addr ((char *) host_name)) == INADDR_NONE)
    {
	host_p = gethostbyname (host_name);

	if (! host_p)
	{
	    fprintf (stderr, "%s: host not found\n", host_name);
	    return (-1);
	}

	(void) memcpy ((char *) (&addr.sin_addr), host_p->h_addr,
			host_p->h_length);
    }

    addr.sin_family  = AF_INET;
    addr.sin_port    = htons ((unsigned short) port);

    if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    {						/* open socket		    */
	perror ("socket");
	return (-1);
    }

    (void) setsockopt (fd, IPPROTO_TCP, TCP_NODELAY,
		       (char *) &opt, sizeof (opt));

    if (connect (fd, (struct sockaddr *) &addr, sizeof (addr)) != 0)
    {
	(void) close (fd);
	perror (host_name);
	return (-1);
    }

    return (fd);
} /* service_connect (char * host_name, int port) */


/*----------------------------------------------------------------------------
 *  main (argc, argv)				- main function
 *----------------------------------------------------------------------------
 */
int
main (int argc, char * argv[])
{
    FILE *  fp;
    char    buf[1024];
    int	    port;
    int	    fd;
    int	    len;

    if (argc != 2)
    {
	fprintf (stderr, "usage: %s message\n", argv[0]);
	exit (1);
    }

    port = 5000;

    fp = fopen ("/var/run/imond.port", "r");

    if (fp)
    {
	(void) fscanf (fp, "%d", &port);
	fclose (fp);
    }

    fd = service_connect ("localhost", port);

    if (fd < 0)
    {
	return (1);
    }

    sprintf (buf, "set-status %s\r\n", argv[1]);
    len = strlen (buf);

    (void) write (fd, buf, len + 2);			    /* send command */
    (void) read (fd, buf, 1024);			    /* read answer  */

    return (0);
} /* main (int argc, char * argv[]) */
