/*
 *		IP_MASQ_MSN msn masquerading module
 *
 *
 * Version:	@(#)ip_masq_msn.c 0.02   06/23/00
 *
 * Author:	Kevin Johnson
 *
 *	Based on Original Work By Wouter Gadeyne
 *		
 *
 * Fixes:
 *      Kevin Johnson           :       Multiple instances of masqs needed
 *
 *
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 *	
 */


#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/init.h>
#include <net/protocol.h>
#include <net/tcp.h>

/* #define IP_MASQ_NDEBUG */
#include <net/ip_masq.h>


/* 
 * Only start with 2 ports, one for TCP and one for UDP
 */

struct ip_masq_app *masq_incarnations[2];

/*
 *	Debug level
 */

#ifdef CONFIG_IP_MASQ_DEBUG
static int debug=0;
MODULE_PARM(debug, "i");
#endif

static int
masq_msn_init_1 (struct ip_masq_app *mapp, struct ip_masq *ms)
{
        MOD_INC_USE_COUNT;
        return 0;
}

static int
masq_msn_done_1 (struct ip_masq_app *mapp, struct ip_masq *ms)
{
        MOD_DEC_USE_COUNT;
        return 0;
}

int
masq_msn_out (struct ip_masq_app *mapp, struct ip_masq *ms, struct sk_buff **skb_p, __u32 maddr)
{
  struct sk_buff *skb;
  struct iphdr *iph;
  struct tcphdr *th;
  struct udphdr *uh;
  unsigned char *data;
  int len,i;
  __u32 from;
  __u16 port;
  struct ip_masq *n_ms;

  skb = *skb_p;
  iph = skb->nh.iph;

  /*
   *  Watch TCP and UDP packets only.  Set the data pointer and xport layer
   *   header pointers where applicable.
   */

  if(iph->protocol==IPPROTO_TCP) {
    th = (struct tcphdr *)&(((char *)iph)[iph->ihl*4]);
    data = (unsigned char *)&(((char *)th)[th->doff*4]);
    len=htons(iph->tot_len)-((iph->ihl<<2)+(th->doff<<2));
    } else if(iph->protocol==IPPROTO_UDP) {
    uh = (struct udphdr *)&(((char *)iph)[iph->ihl*4]);
    data = (unsigned char *)&(((char *)uh)[8]); // 8 bytes long for UDP
    len=htons(iph->tot_len)-((iph->ihl<<2)+8);
    } else return 0;

  /*
   *  All the packets we have to monitor are bofa (B0FA) packets
   */

  if((len<20) || (data[2]!=0xB0) || (data[3]!=0xFA)) return 0;


  /*
   * *sigh*
   * we must go through this packet looking for the 02 00 byte sequence,
   *  and *sigh-squared* unfortunately, it may not be word aligned...
   *
   * The length of the packet is the first 2 bytes. (big-endian)
   */

  len=htons(* (short *) (data));

  for(i=4;i<len;++i) {

    /*
     *  All ports will start with 02 00 key identifing them
     */ 

    if(*(short *) (data+i)==2) {

      port=htons(* (short *) (data+i+2));

      /*
       * because I do not know the packet format, I cannot use data block
       * sizes to skip over text bytes, so lets check to see that we didn't
       * find something other than a direct play port (2300-2400)
       */

      if(port>=2300 && port<=2400) {

        from = ntohl(ms->saddr);

	IP_MASQ_DEBUG(1-debug, "PORT %X:%X detected\n",from,port);

	/*
	 * Now update or create a masquerade entry for it
	 */

	IP_MASQ_DEBUG(1-debug, "protocol %d %lX:%X %X:%X\n", iph->protocol, htonl(from), htons(port), iph->daddr, 0);

	n_ms = ip_masq_out_get(IPPROTO_TCP,
			       htonl(from), htons(port),
			       iph->daddr, 0);
        if (!n_ms) {
          n_ms = ip_masq_new(IPPROTO_TCP,
			     maddr, 0,
		  	     htonl(from), htons(port),
			     iph->daddr, 0,
			     IP_MASQ_F_NO_DPORT);

	  /* sanity check--it ought not be null */

	  if (n_ms==NULL) return 0;

          /*
 	   * This new TCP port must be monitored just like the one we
           *  are monitoring now.  Since we don't know the destination
           *  port, we simply do our own bind here.
           * Just in case the port has a binding registered (we do not
           * want to use it), we unbind it.  Then we do our own bind
           * making sure we call masq_init_1, because when the port
           * expires, it will call masq_done_1.
           */


          if(n_ms->app && n_ms->app->masq_done_1) 
            n_ms->app->masq_done_1(n_ms->app,n_ms);
          n_ms->app=mapp;
          mapp->n_attach++;
          if(mapp->masq_init_1) 
            mapp->masq_init_1(mapp, n_ms);

	  ip_masq_control_add(n_ms, ms);

	  }

	/*
         * make sure we are listening (on the old masq entry or the new one)
         *  and reset the expire timer
         */

	ip_masq_listen(n_ms);
	n_ms->timeout = 30*HZ;
	ip_masq_put(n_ms);

	/*
	 * Replace the old PORT with the new one in the message
	 */

	from = ntohl(n_ms->maddr);
	port = ntohs(n_ms->mport);

	IP_MASQ_DEBUG(1-debug, "new PORT %X:%X\n",from,port);

        * (short *) (data+i+2)=htons(port);

	/*
	 * skip 16 bytes, 15 here, 1 by loop increment
	 */

        i+=15;

	/* 
         * A UDP port might follow 16 bytes after TCP port
         */ 

        if((i+1<len)
           && (*(short *) (data+i+1)==2)) {

          port=htons(* (short *) (data+i+3));

          /*
           * again, check to make sure the port is 2300-2400
           */

          if((port>=2300) && (port<=2400)) {

	    /* 
             * add that extra byte the loop counter was going to add
             */

 	    ++i; 

	    from = ntohl(ms->saddr);

	    IP_MASQ_DEBUG(1-debug, "PORT %X:%X detected\n",from,port);

	    /*
	     * Now update or create a UDP masquerade entry for it
	     */

	    IP_MASQ_DEBUG(1-debug, "protocol %d %lX:%X %X:%X\n", iph->protocol, htonl(from), htons(port), iph->daddr, 0);

	    n_ms = ip_masq_out_get(IPPROTO_UDP,
				 htonl(from), htons(port),
				 iph->daddr, 0);
            if (!n_ms) {
              n_ms = ip_masq_new(IPPROTO_UDP,
				 maddr, 0,
				 htonl(from), htons(port),
				 iph->daddr, 0,
				 IP_MASQ_F_NO_DPORT);

              /*
               * Yet another sanity check
               */

 	      if (n_ms==NULL) return 0;

	      /*
               *  Adding UDP masq entries is easier
               */

              ip_masq_control_add(n_ms, ms);

	      }

	    /* 
             * update the expire time
             */

	    n_ms->timeout = 5*HZ;
	    ip_masq_put(n_ms);

	    /*
	     * Replace the old PORT with the new one
	     */

	    from = ntohl(n_ms->maddr);
	    port = ntohs(n_ms->mport);

	    IP_MASQ_DEBUG(1-debug, "new PORT %X:%X\n",from,port);

            * (short *) (data+i+2)=htons(port);

	    /*
	     * skip 16 bytes (15 here, loop counter will skip the last byte)
             *  (this time for sure!)
	     */

	    i+=15;

            }
          }
        }
      }
    }

  return 0;
  }

int
masq_msn_in (struct ip_masq_app *mapp, struct ip_masq *ms, struct sk_buff **skb_p, __u32 maddr)
{
  struct sk_buff *skb;
  struct iphdr *iph;
  struct tcphdr *th;
  unsigned char *data;
  __u32 from;
  __u16 port;
  struct ip_masq *n_ms;

  skb = *skb_p;
  iph = skb->nh.iph;

  if(iph->protocol==IPPROTO_TCP) {
    th = (struct tcphdr *)&(((char *)iph)[iph->ihl*4]);
    data = (unsigned char *)&(((char *)th)[th->doff*4]);
    } else return 0;

	//len=iph->len-((iph->ihl<<2)+(th->diff<<2));

  if(th->ack_seq==0) {
    n_ms = ip_masq_new(IPPROTO_TCP,
		       ms->maddr, ms->mport,
		       ms->saddr,ms->sport,
		       ms->daddr,0,
		       IP_MASQ_F_NO_DPORT);
          

    } else {
	  
    if((data[2]!=0xB0) || (data[3]!=0xFA)) return 0;

    port=htons(* (short *) (data+6));
    from = ntohl(ms->saddr);

    IP_MASQ_DEBUG(1-debug, "PORT %X:%X detected\n",from,port);

    /*
     * we expect the client to make a TCP connection to this port to the
     * same machine we got this message from
     */

    IP_MASQ_DEBUG(1-debug, "protocol %d %lX:%X %X:%X\n", iph->protocol, htonl(from), htons(port), iph->daddr, 0);

    n_ms = ip_masq_out_get(IPPROTO_TCP,
			   from, 0,
			   iph->saddr, htons(port));

    if(!n_ms) {
      n_ms = ip_masq_new(IPPROTO_TCP,
			 maddr, 0,
			 htonl(from), 0,
			 iph->saddr, htons(port),
			 IP_MASQ_F_NO_SPORT);

      }
    }

  /*
   * This new TCP port must be monitored just like the one we
   *  are monitoring now.  Since we don't know the destination
   *  port, we simply do our own bind here.
   * Just in case the port has a binding registered (we do not
   * want to use it), we unbind it.  Then we do our own bind
   * making sure we call masq_init_1, because when the port
   * expires, it will call masq_done_1.
   */

  if (n_ms==NULL) return 0;

  if(n_ms->app && n_ms->app->masq_done_1) n_ms->app->masq_done_1(n_ms->app,n_ms);
  n_ms->app=mapp;
  mapp->n_attach++;
  if(mapp->masq_init_1) mapp->masq_init_1(mapp, n_ms);

  ip_masq_control_add(n_ms, ms);
  ip_masq_listen(n_ms);

  n_ms->timeout = 30*HZ;
  ip_masq_put(n_ms);

  return 0;
  }


struct ip_masq_app ip_masq_msn = {
        NULL,			/* next */
	"msn",			/* name */
        0,                      /* type */
        0,                      /* n_attach */
        masq_msn_init_1,        /* ip_masq_init_1 */
        masq_msn_done_1,        /* ip_masq_done_1 */
        masq_msn_out,           /* pkt_out */
        masq_msn_in,            /* pkt_in (would be masq_msn_in) */
};

/*
 * 	ip_masq_msn initialization
 */

__initfunc(int ip_masq_msn_init(void))
{
	int j;

	if ((masq_incarnations[0] = kmalloc(sizeof(struct ip_masq_app), GFP_KERNEL)) == NULL) return -ENOMEM;
	memcpy(masq_incarnations[0], &ip_masq_msn, sizeof(struct ip_masq_app));
	if ((j = register_ip_masq_app(masq_incarnations[0], IPPROTO_TCP, 47624))) return j;
	if ((masq_incarnations[1] = kmalloc(sizeof(struct ip_masq_app), GFP_KERNEL)) == NULL) return -ENOMEM;
	memcpy(masq_incarnations[1], &ip_masq_msn, sizeof(struct ip_masq_app));
	if ((j = register_ip_masq_app(masq_incarnations[1], IPPROTO_UDP, 47624))) return j;
	IP_MASQ_DEBUG(1-debug, "Msn: loaded support on port[0 and 1] = %d\n", 1, 47624);
	return 0;
}

/*
 * 	ip_masq_msn fin.
 */

int ip_masq_msn_done(void)
{
	int j, k;

	k=0;
	if ((j = unregister_ip_masq_app(masq_incarnations[0]))) {
		k = j;
		} else {
		kfree(masq_incarnations[0]);
		masq_incarnations[0] = NULL;
		IP_MASQ_DEBUG(1-debug, "Msn: unloaded support on port[%d] = %d\n", 0, 47624);
		}
	if ((j = unregister_ip_masq_app(masq_incarnations[1]))) {
		k = j;
		} else {
		kfree(masq_incarnations[1]);
		masq_incarnations[1] = NULL;
		IP_MASQ_DEBUG(1-debug, "Msn: unloaded support on port[%d] = %d\n", 1, 47624);
		}
	return k;
}

#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
        if (ip_masq_msn_init() != 0)
                return -EIO;
        return 0;
}

void cleanup_module(void)
{
        if (ip_masq_msn_done() != 0)
                printk(KERN_INFO "ip_masq_msn: can't remove module");
}

#endif /* MODULE */
