/*  gtk-imonc - imond client for fli4l
 *  Copyright (C) 2001 Stefan Strigler <zeank@x-berg.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "imonc.h"
#include "callbacks.h"
#include "timer.h"

void show_usage(char *argv[])
{
	fprintf (stderr, "usage: %s [-n] [-m] [-p pass] hostname [port]\n", argv[0]);
	fprintf (stderr, "  whereas:\n");
	fprintf (stderr, "  -n\t\tdon't display news panel\n");
	fprintf (stderr, "  -m\t\tstart in minimode\n");
	fprintf (stderr, "  -p pass\tuse pass as password\n");
	exit(1);
}

int
main (int argc, char *argv[])
{
	char *answer;
	gchar buf[64];
	gchar pass[64];
	struct sigaction act;
	guint arg = 1;

	port = IMOND_PORT;

	siginterrupt (SIGALRM, 1);

	/* Assign sig_pipe as SIGPIPE handler */
	act.sa_handler = fli4l_reconnect;

	/* We don't want to block any other signals */
	sigemptyset(&act.sa_mask);

	if (sigaction(SIGPIPE, &act, NULL) < 0) {
		fprintf(stderr, "sigaction failed\n");
		return 1;
	}

	signal (SIGALRM, fli4l_reconnect);

#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
	textdomain (PACKAGE);
#endif

	gtk_set_locale ();
	gtk_init (&argc, &argv);

	add_pixmap_directory (PACKAGE_DATA_DIR "/pixmaps");
	add_pixmap_directory (PACKAGE_SOURCE_DIR "/pixmaps");

	/* parse command line arguments
	 * proposed format:
	 * $ gtk-imonc [-n] [-p pass] hostname [port]
	 */

	if (argc < 2) 
		show_usage(argv);

	use_news_tab = TRUE;
	if (strcmp (argv[arg], "-n") == 0) {
		use_news_tab = FALSE;
		arg++;
	}

	is_mini = FALSE;
	if (strcmp (argv[arg], "-m") == 0) {
		is_mini = TRUE;
		arg++;
	}

	if (argc<=arg)
		show_usage(argv);

	memset(pass,'\0',64);
	if (strcmp (argv[arg], "-p") == 0) {
		arg++;
		if (argc <= arg)
			show_usage(argv);
		strncpy(pass,argv[arg],63);
		memset(argv[arg],'x',strlen(argv[arg])); /* hide password */
		arg++;
	}

	if (argc <= arg)
		show_usage(argv);
	
	host_name = argv[arg++];
	
	if (arg < argc)
		port = atoi (argv[arg]);
	
	if ((fd = service_connect (host_name, port)) < 0) {
		exit (1);
	}


/* 	halt_warning = create_halt_warning (); */
/* 	reboot_warning = create_reboot_warning (); */

	gtk_imonc = create_gtk_imonc ();
	gtk_miniimonc = create_gtk_miniimonc();
	rightclick_menu = create_rightclick_menu();
	
	/* check password */
	if (strlen(pass)>0) {
		sprintf (buf, "pass %s", pass);
		send_command (fd, buf);
	} else {
		send_command (fd, "pass");
	}

	is_admin = FALSE; /* remember if we are in admin mode */
	answer = get_answer (fd);
			
	/* 	From Nico Wallmeier:
	 *		1: Userpasswort ntig
	 * 		2: Adminpasswort ntig
	 * 		3: User+ Adminpasswort ntig
	 * 		+4 im Adminmodus 
	 */

	if (!strlen(answer) || /* supplied password was wrong */
			atoi(answer) == 1 || atoi(answer) == 3) { 
		/* need to check password */
		GtkWidget *auth_dialog;
		
		auth_dialog = create_auth_dialog ();
		gtk_widget_show (auth_dialog);
	} else {
		/* check if we are in admin mode */
		if (atoi (answer)>=4)
			is_admin = TRUE; /* we are already in admin mode */

		gtk_widget_show (gtk_imonc);
		imonc_init (gtk_imonc); /* initialize imonc functionality */
		imonc_update (gtk_imonc); /* init with current values */
		

		if (is_mini) {
			GtkWidget *mini_view;

			is_mini = FALSE;
			mini_view = lookup_widget (rightclick_menu,"mini_view");
			gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mini_view),TRUE);
		}
		timeout_tag = gtk_timeout_add (TIMEOUT_MSEC, imonc_update, gtk_imonc);
	}
	
	gtk_main ();
	
	service_disconnect (fd);
	return 0;
}
